"""File providing some score functions to be used in optimization."""

import numpy as np


def MAD(tdata, data):
    """Scorefunction mean absolute difference.

    Parameters
    ----------
    tdata : np.array
        Data to be fitted against simulation output.
    data : np.array
        Simulation output

    Returns
    -------
    value : float
        Score of Mean Absolute Difference.
    """
    abs_diff = np.abs(tdata - data)
    return abs_diff[~np.isnan(abs_diff)].mean()


def NormalizedMAD(tdata, data):
    """Scorefunction normalized mean absolute difference.

    Parameters
    ----------
    tdata : np.array
        Data to be fitted against simulation output.
    data : np.array
        Simulation output

    Returns
    -------
    value : float
        Score of Normalized Mean Absolute Difference.

    """
    abs_diff = np.abs(tdata - data) / tdata
    return abs_diff[~np.isnan(abs_diff)].mean()



"""All functionality pertaining to distribution analysis."""

import numpy as np
import scipy.stats.kde as kde


def hpd_grid(sample, alpha=0.05, roundto=2):
    """Calculate highest posterior density (HPD) of array for given alpha.

    The HPD is the minimum width Bayesian credible interval (BCI).
    The function works for multimodal distributions, returning more than one mode

    Parameters
    ----------
    sample : Numpy array or python list
        An array containing MCMC samples
    alpha : float
        Desired probability of type I error (defaults to 0.05)
    roundto : integer
        Number of digits after the decimal point for the results (Default value = 2)

    Returns
    -------
    hpd : np.array
        Highest Posterior Density
    x : list()
        x-values of the Kernel Density Estimation (KDE)
    y : list()
        y-values of the KDE
    modes : list()
        List of locations from the modes of the distribution. Usually one
        location, but also finds 2 if bimodal.
    """
    sample = np.asarray(sample)
    sample = sample[~np.isnan(sample)]
    # get upper and lower bounds
    lb = np.min(sample)
    ub = np.max(sample)
    density = kde.gaussian_kde(sample)
    x = np.linspace(lb, ub, 2000)
    y = density.evaluate(x)
    # y = density.evaluate(x, l, u) waitting for PR to be accepted
    xy_zipped = zip(x, y / np.sum(y))
    xy = sorted(xy_zipped, key=lambda x: x[1], reverse=True)
    xy_cum_sum = 0
    hdv = []
    for val in xy:
        xy_cum_sum += val[1]
        hdv.append(val[0])
        if xy_cum_sum >= (1 - alpha):
            break
    hdv.sort()
    diff = (ub - lb) / 20  # differences of 5%
    hpd = []
    hpd.append(round(min(hdv), roundto))
    for i in range(1, len(hdv)):
        if hdv[i] - hdv[i - 1] >= diff:
            hpd.append(round(hdv[i - 1], roundto))
            hpd.append(round(hdv[i], roundto))
    hpd.append(round(max(hdv), roundto))
    ite = iter(hpd)
    hpd = list(zip(ite, ite))
    modes = []
    for value in hpd:
        x_hpd = x[(x > value[0]) & (x < value[1])]
        y_hpd = y[(x > value[0]) & (x < value[1])]
        modes.append(round(x_hpd[np.argmax(y_hpd)], roundto))
    return hpd, x, y, modes



"""Visualization unit of network data."""

from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.collections import LineCollection
import matplotlib.animation as animation
import matplotlib.ticker as ticker
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np


def plot_grid(ax, model, layout, with_labels=False, title=''):
    """Plot grid network.

    Parameters
    ----------
    ax : matplotlib.axis
        Matplotlib axis to plot in.
    model : Mesa.model
        Mesa agent-based model object.
    layout : nx.__layout
        Networkx layout.
        (e.g., nx.kamada_kawai_layout, nx.spring_layout)
    with_labels : boolean
        True or false boolean on whether or not to show labels.
        (Default value = False)
    title : str
        Title of the plot. (Default value = '')

    Returns
    -------
    None

    """
    cmap = plt.get_cmap("viridis")
    graph = model.G
    pos = layout(graph)

    states = [int(i.k) for i in model.grid.get_all_cell_contents()] # FIX
    colors = [cmap(30 + i * 30) for i in states]
    sizes = [30 + i * 30 for i in states]
    nx.draw(
        graph,
        pos,
        node_size=sizes,
        edge_color='gray',
        node_color=colors,
        with_labels=with_labels,
        alpha=0.9,
        font_size=14,
        ax=ax)
    ax.set_title(title)
    return


def nxNetworkMP4(full_graph, Gs, labels, layout, save_path, remove_inactive_nodes=False):
    """Create an MP4 from the network dynamics over time.

    Parameters
    ----------
    full_graph : nx.Graph
        Networkx graph with all nodes and edges.
    Gs : List[nx.Graph
        List of networkx graphs that change over time.
    labels : List[str
        List of labels showing the current time.
    layout : nx.__layout
        Networkx layout.
        (e.g., nx.kamada_kawai_layout, nx.spring_layout)
    save_path : str
        Where do you want to save your mp4?
    remove_inactive_nodes : bool
        True to remove inactive nodes.
        Otherwise full nodelist is always visible. Defaults to False.

    Returns
    -------
    None

    """
    assert len(Gs) == len(labels), "Desired (len(Gs) == len(labels))"

    pos = layout(full_graph)
    fig, ax = plt.subplots(figsize=(15, 15))
    ax.axis("off")

    # Add Edges
    segments = np.array([[pos[x], pos[y]] for x, y in Gs[0].edges()])
    edges = LineCollection(segments, alpha=0.4, linewidths=2, color="#727272")
    ax.add_artist(edges)

    # Add Nodes
    coordinates = np.array([i for i in pos.values()]).T
    nodes = ax.scatter(*coordinates, s=8, alpha=1, color="#ba0000")

    # Add Time indication
    text = ax.text(1.1, 1, labels[0], ha="right", fontsize=56, color='C1', wrap=True)

    # Combine elements into list to pass later on
    actors = [edges, nodes, text]

    def update(i: int, edges, nodes, text, pos, ax):
        """

        Parameters
        ----------
        i: int
            Index of frames.
        edges : np.array
            List of edge locations of previous step.
        nodes : np.array
            List of node locations of previous step.
        text : str
            Text to be shown in upper right corner at each update.
        pos : dict
            Networkx - Matplotlib positional layout dictionary.
        ax : matplotlib.axis
            Matplotlib axis to plot in.


        Returns
        -------
        None

        """
        segments = [[pos[x], pos[y]] for x, y in Gs[i].edges()]
        edges.set_paths(segments)
        if remove_inactive_nodes:
            points = [pos[node] for node in Gs[i].nodes()]
            nodes.set_offsets(points)
        text.set_text(labels[i])

    # Create animation and save it
    ani = animation.FuncAnimation(fig,
                                  update,
                                  frames=len(Gs),
                                  fargs=(*actors, pos, ax),
                                  interval=350,
                                  repeat=False)
    ani.save(save_path)


def heatmap(data, row_labels, col_labels, ax=None, fig=None,
            cbar_kw={}, cbarlabel="", title="", **kwargs):
    """Create a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    data : np.array(
        A 2D numpy array of shape (N, M).
    row_labels : List[str]
        A list or array of length N with the labels for the rows.
    col_labels : List[str]
        A list or array of length M with the labels for the columns.
    ax : matplotlib.axes.Axes
        A `matplotlib.axes.Axes` instance to which the
        heatmap is plotted. If not provided, use current axes or create a new one. Defaults to None.
    fig : matplotlib.figure
        Matplotlib figure object. Defaults to None.
    cbar_kw : dict
        A dictionary with arguments to `matplotlib.Figure.colorbar`. Defaults to {}.
    cbarlabel : str
        The label for the colorbar. Defaults to "".
    title : str
        Title for subfigure. Defaults to "".
    **kwargs : dict
        All other arguments are forwarded to `imshow`.

    Returns
    -------
    im : matplotlib.axis.imshow
        The heatmap object.
    cb : fig.colorbar
        The matplotlib colorbar object as a color legend.

    """
    if not ax:
        ax = plt.gca()

    # Set title
    ax.set_title(title, size=15, pad=20)

    # Plot the heatmap
    im = ax.imshow(data, **kwargs)

    # Create colorbar
    ax_divider = make_axes_locatable(ax)
    # Add an axes to the right of the main axes.
    cax = ax_divider.append_axes("right", size="5%", pad="2%")
    cb = fig.colorbar(im, cax=cax)
    cb.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

    # We want to show all ticks...
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels)
    ax.set_yticklabels(row_labels)

    # Let the horizontal axes labeling appear on top.
    ax.tick_params(top=True, bottom=False,
                   labeltop=True, labelbottom=False)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=0, ha="right",
             rotation_mode="anchor")

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_yticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Turn spines off and create white grid.
    ax.spines[:].set_visible(False)

    ax.set_xticks(np.arange(data.shape[1]+1)-.5, minor=True)
    ax.set_yticks(np.arange(data.shape[0]+1)-.5, minor=True)
    ax.grid(which="minor", color="w", linestyle='-', linewidth=3)
    ax.tick_params(which="minor", bottom=False, left=False)

    return im, cb


def annotate_heatmap(im, data=None, valfmt="{x:.2d}",
                     textcolors=("black", "white"),
                     threshold=None, **textkw):
    """Annotate a heatmap.

    Source:
        https://matplotlib.org/stable/gallery/images_contours_and_fields/image_annotated_heatmap.html

    Parameters
    ----------
    im : AxesImage
        The AxesImage to be labeled.
    data : List[str]
        Data used to annotate.  If None, the image's
        data is used. Defaults to None.
    valfmt : str
        The format of the annotations inside the heatmap.
        This should either use the string format method, e.g. "$ {x:.2f}",
        or be a `matplotlib.ticker.Formatter`. Defaults to "{x:.2d}".
    textcolors : tuple
        A pair of colors.  The first is used for values
        below a threshold, the second for those above. Defaults to ("black", "white").
    threshold : float
        Value in data units according to which the colors from
        textcolors are applied.  If None (the default) uses the middle of the colormap as
        separation. Defaults to None.
    "white") :

    **textkw :


    Returns
    -------
    texts : List[str]
        The texts that go in each cell of the heatmap.

    """
    if not isinstance(data, (list, np.ndarray)):
        data = im.get_array()

    # Normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)
    else:
        threshold = im.norm(data.max())/2.

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw = dict(horizontalalignment="center",
              verticalalignment="center")
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = ticker.StrMethodFormatter(valfmt)

    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
            text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)

    return texts


def plotHeatmaps(gs, dm, cmaps, suptitle="Degree measures over time",
                 path="centralities.png"):
    """Plot network heatmap centrality mappings changing over time.

    Parameters
    ----------
    gs : List[nx.Graph]
        List of networkx graphs over time.
    dm : List[List[nx.__centrality function]]
        2d list of centrality measures.
        Shape will be same as grid for heatmap visualizations
    cmaps : List[List[matplotlib colormaps]]
        2d list of same shape as dm.
    suptitle :
         (Default value = "Degree measures over time")
    path :
         (Default value = "centralities.png")

    Returns
    -------
    None

    """
    fig, axes = plt.subplots(dm.shape[0], dm.shape[1], figsize=(30, 13.5))
    fig.suptitle(suptitle, fontsize=20)
    for i, axs in enumerate(axes):
        for j, ax in enumerate(axs):
            # Get degree data
            try:
                degrees = [dm[i][j](g) for g in gs]
            except nx.exception.PowerIterationFailedConvergence:
                degrees = [dm[i][j](g, max_iter=500) for g in gs]

            # Find max degree value
            max_ = max([max(list(degree.values())) for degree in degrees])

            # Create histogram for each timestep
            freqs = [
                np.histogram(list(degree.values()), bins=np.linspace(0, max_,
                                                                     10))[0]
                for degree in degrees
            ]
            # Normalize frequencies
            freqs = np.array([f / sum(f) for f in freqs])

            # Create datastruct
            data = {
                "timepoints": list(range(len(freqs))),
                "bins": ['{0:.1e}'.format(t) for t in np.linspace(0, max_, 10)][:-1],
                "data": freqs.T
            }

            # Make heatmap
            im, cbar = heatmap(data["data"],
                               data["bins"],
                               data["timepoints"],
                               ax=ax,
                               fig=fig,
                               cmap=cmaps[i][j],
                               cbarlabel="Frequency in %",
                               title=f"{dm[i][j].__name__}")
            annotate_heatmap(im, valfmt="{x:.2f}")

    fig.tight_layout()
    plt.savefig(path)
    plt.show()



"""In this file we hold all useful decorators that we develop over time.

For hints check:
https://betterprogramming.pub/python-decorators-5-advanced-features-to-know-17dd9be7517b
"""

from functools import wraps
import time


def logging_time(unit):
    """Decorate inner function and log time.

    Parameters
    ----------
    func : Function
        Decorate function such that it prints the elapsed runtime.
    unit : str
        Time constant in which you want to see how long the operation took.

    Returns
    -------
    logger : Inner function of the wrapper.

    """
    def logger(func):
        @wraps(func)
        def inner_logger(*args, **kwargs):
            """Log time."""
            start = time.time()
            func(*args, **kwargs)
            if unit == "ms":
                scaling = 1000
            elif unit == "m":
                scaling = 1/60
            else:
                scaling = 1
            print(f"Calling {func.__name__}: {(time.time() - start) * scaling:.5f} {unit}")

        return inner_logger

    return logger


if __name__ == "__main__":

    @logging_time("s")
    def sleep():
        """Sleep."""
        time.sleep(3)

    @logging_time("ms")
    def sleep2():
        """Sleep."""
        time.sleep(2.1)

    print("sleep")
    sleep()
    print("sleep2")
    sleep2()



"""__init__.py initiation file."""

from .analyse import *
from .datastructures import *
from .decorators import *
from .modeling import *
from .optimize import *
from .processdata import *
from .visualize import *



"""Own functions to process data."""

import numpy as np


def perturb(data, pmean, pstd, prem):
    """Perturb the data.

    Parameters
    ----------
    data : np.array
        Data that you'd want to perturb.
    pmean : float
        Mean of the normal distribution that perturbs the data.
    pstd : float
        Standard deviation of normal distribution that perturbs the data.
    prem : float
        Ratio of datapoints that are to be removed randomly [0,1].

    Returns
    -------
    data : np.array
        Perturbed data.

    """
    # Perturb data
    data *= np.random.normal(pmean, pstd, data.shape)

    # Choose certain ratio of the data by random to illustrate loss of data
    data *= np.random.choice([np.nan, 1], size=(data.shape), p=[prem, 1 - prem])

    return data


def splitData(data, fraction_train=2 / 3, type_split="time"):
    """Split data into seen (training set) and unseen (test set) data.

    Returns two sets of equal dimensions with NaN at the places where
    there is no data in that set. The two are complimentary.

    Parameters
    ----------
    data : np.array
        Data that you'd want to split.
    fraction_train : float [0,1]
        Fraction of data that is used as training data. (Default value = 2 / 3)
    type_split : str
        How to split the data, randomly or over time. Possible answers are:
        ["random", "time"]. (Default value = "time")

    Returns
    -------
    train : np.array
        Training set.
    test : np.array
        Test set.

    """
    # Get a list of data indices
    nonnans = np.argwhere(~np.isnan(data))
    index = np.arange(len(nonnans))

    # Copy original data structure to not mutate it
    train = data.copy()
    test = data.copy()

    # Specify which type of split we want
    # Split randomly over all datapoints
    if type_split == "random":
        # Shuffle the data in a random order
        np.random.shuffle(index)

        # Make division of data
        div = int(np.floor(len(index) * fraction_train))
        test_nan = index[:div]
        train_nan = index[div:]

        # Rebuild data into the two sets
        for i in train_nan:
            ti = nonnans[i]
            train[ti[0]][ti[1]] = np.nan
        for i in test_nan:
            ti = nonnans[i]
            test[ti[0]][ti[1]] = np.nan

    # Split over a section in time
    elif type_split == "time":
        # Make division of data
        div = int(np.floor(len(index) * fraction_train))
        test_nan = index[:div]
        train_nan = index[div:]

        # Rebuild data into the two sets
        for i in train_nan:
            ti = nonnans[i]
            train[ti[0]][ti[1]] = np.nan
        for i in test_nan:
            ti = nonnans[i]
            test[ti[0]][ti[1]] = np.nan

    return train, test



"""All commonly used pydantic questionnaire dataclasses ready for reuse."""

from typing import Literal
from decimal import Decimal
from datetime import datetime

from pydantic import BaseModel, validator, conint, condecimal


class Wellbeing(BaseModel):
    """Questionnaire regarding to the wellbeing of the participant."""

    q_start: datetime
    q_end: datetime
    well_0: conint(ge=0, le=3)
    well_1: conint(ge=0, le=3)
    well_2: conint(ge=0, le=3)
    well_3: conint(ge=0, le=3)
    well_4: conint(ge=0, le=3)
    well_5: conint(ge=0, le=3)
    well_6: conint(ge=0, le=3)
    well_7: conint(ge=0, le=3)
    well_8: conint(ge=0, le=3)
    well_9: conint(ge=0, le=3)
    well_10: conint(ge=0, le=3)
    well_11: conint(ge=0, le=3)
    well_12: conint(ge=0, le=3)
    well_13: conint(ge=0, le=3)
    well_14: conint(ge=0, le=3)

    @property
    def score(self):
        """Calculate score of questionnaire."""
        return sum([getattr(self, "well_" + str(i)) for i in range(15)]) / 45


class Mobility(BaseModel):
    """Short questionnaire regarding mobility of participant."""

    q_start: datetime
    q_end: datetime
    mob0: conint(ge=0, le=2)
    mob_freq0: conint(ge=0, le=3)
    mob_indep0: conint(ge=0, le=2)
    mob1: conint(ge=0, le=2)
    mob_freq1: conint(ge=0, le=3)
    mob_indep1: conint(ge=0, le=2)
    mob2: conint(ge=0, le=2)
    mob_freq2: conint(ge=0, le=3)
    mob_indep2: conint(ge=0, le=2)
    mob3: conint(ge=0, le=2)
    mob_freq3: conint(ge=0, le=3)
    mob_indep3: conint(ge=0, le=2)
    mob4: conint(ge=0, le=2)
    mob_freq4: conint(ge=0, le=3)
    mob_indep4: conint(ge=0, le=2)

    @property
    def score(self):
        """Calculate score of questionnaire."""
        m_map = {0: 1, 1: 0.5, 2: 0}
        mi_map = {0: 1, 1: 1.5, 2: 2}

        m = [getattr(self, "mob" + str(i)) for i in range(5)]
        mf = [getattr(self, "mob_freq" + str(i)) for i in range(5)]
        mi = [getattr(self, "mob_indep" + str(i)) for i in range(5)]

        return sum([(i + 1) * m_map[m[i]] * (mf[i] + 1) * mi_map[mi[i]]
                     for i in range(5)])


class Loneliness(BaseModel):
    """Loneliness questionnaire answers.
    
    Data is structures in a few categories to find them easier, the following prefixes are defined:
    
    q => metadata about questionnaire (e.g., timestamps)
    
    l => loneliness questions

    Parameters
    ----------
    BaseModel : pydantic.BaseModel
        Pydantic base dataclass

    Returns
    -------

    """

    q_start: datetime
    q_end: datetime
    l_0: conint(ge=0, le=2)
    l_1: conint(ge=0, le=2)
    l_2: conint(ge=0, le=2)
    l_3: conint(ge=0, le=2)
    l_4: conint(ge=0, le=2)
    l_5: conint(ge=0, le=2)
    l_6: conint(ge=0, le=2)
    l_7: conint(ge=0, le=2)
    l_8: conint(ge=0, le=2)
    l_9: conint(ge=0, le=2)
    l_10: conint(ge=0, le=2)

    @property
    def emo_l(self):
        """Calculate emotional loneliness score."""
        emotional_items = [1, 2, 4, 5, 8, 9]
        return sum([
            1 for q in emotional_items
            if getattr(self, "l_" + str(q)) in [0, 1]
        ])

    @property
    def soc_l(self):
        """Calculate social loneliness score."""
        social_items = [0, 3, 6, 7, 10]
        return sum([
            1 for q in social_items if getattr(self, "l_" + str(q)) in [1, 2]
        ])

    @property
    def tot_l(self):
        """Calculate total loneliness score."""
        return self.emo_l + self.soc_l

    @validator('q_start', "q_end", pre=True)
    def time_validate(cls, v):
        """Make sure that the date type formatting works out.

        Parameters
        ----------
        v :
            

        Returns
        -------

        """
        return datetime.fromisoformat(v)


class Topics(BaseModel):
    """Topics questionnaire answers.
    
    Data is structures in a few categories to find them easier the following prefixes are defined:
    
    q => metadata about questionnaire (e.g., timestamps)
    
    v => questions about your feeling (voelen)
    
    l => questions about physical attributes (lichamelijk)
    
    p => questions about practical activities (praktisch)
    
    m => questions about the mouth (mond)
    
    z => questions about diseases (ziekte)

    Parameters
    ----------
    BaseModel : pydantic.BaseModel
        Pydantic base dataclass

    Returns
    -------

    """

    q_start: datetime
    q_end: datetime
    s_soc_act: condecimal(multiple_of=Decimal('0.2'), le=1)
    v_zenuw: condecimal(multiple_of=Decimal('0.2'), le=1)
    v_somber: condecimal(multiple_of=Decimal('0.2'), le=1)
    v_put: condecimal(multiple_of=Decimal('0.2'), le=1)
    v_kalm: condecimal(multiple_of=Decimal('0.2'), le=1)
    v_gelukkig: condecimal(multiple_of=Decimal('0.2'), le=1)
    v_pijn: condecimal(multiple_of=Decimal('0.2'), le=1)
    v_angst: condecimal(multiple_of=Decimal('0.2'), le=1)
    v_kw_leven: condecimal(multiple_of=Decimal('0.1'), le=1)
    l_lopen: condecimal(multiple_of=Decimal('0.2'), le=1)
    l_geheugen: conint(ge=0, le=1)
    l_gevallen: conint(ge=0, le=1)
    p_wassen_aankleden: condecimal(multiple_of=Decimal('0.2'), le=1)
    p_dagelijkse_act: condecimal(multiple_of=Decimal('0.2'), le=1)
    p_kleden: Literal[0, 0.33, 0.66, 1]
    p_stoel: Literal[0, 0.33, 0.66, 1]
    p_afdrogen: Literal[0, 0.33, 0.66, 1]
    p_trap: Literal[0, 0.33, 0.66, 1]
    p_rondlopen: Literal[0, 0.33, 0.66, 1]
    p_voet: Literal[0, 0.33, 0.66, 1]
    p_huishouden: Literal[0, 0.33, 0.66, 1]
    p_boodschap: Literal[0, 0.33, 0.66, 1]
    p_medicijnen: Literal[0, 0.33, 0.66, 1]
    p_vervoer: Literal[0, 0.33, 0.66, 1]
    m_pijn: Literal[0, 0.25]
    m_kauw: Literal[0, 0.25]
    m_droog: Literal[0, 0.25]
    m_slik: Literal[0, 0.25]
    z_suikerziekte: Literal[0, 1]
    z_beroerte: Literal[0, 1]
    z_hartfalen: Literal[0, 1]
    z_kanker: Literal[0, 1]
    z_astma: Literal[0, 1]
    z_urineverlies: Literal[0, 1]
    z_gewrichtsslijtage: Literal[0, 1]
    z_gewrichtsontsteking: Literal[0, 1]
    z_botontkalking: Literal[0, 1]
    z_botbreuken: Literal[0, 1]
    z_duizeligheid: Literal[0, 1]
    z_parkinson: Literal[0, 1]
    z_depressie: Literal[0, 1]
    z_paniekstoornis: Literal[0, 1]
    z_dementie: Literal[0, 1]
    z_gehoorproblemen: Literal[0, 1]
    z_zien: Literal[0, 1]

    @property
    def frailty_index(self):
        """Calculate frailty based on TOPICS questionnaire."""
        return round(
            sum([float(v)
                 for k, v, in self.__dict__.items() if k[0] != "q"]) / 42.0, 3)

    @validator('q_start', "q_end", pre=True)
    def time_validate(cls, v):
        """Make sure that the date type formatting works out.

        Parameters
        ----------
        v :
            

        Returns
        -------

        """
        return datetime.fromisoformat(v)



"""For all your time handeling needs, here it is."""

from datetime import datetime
import time
import re


def extract_date(date):
    """Make datetime object from string.

    Parameters
    ----------
    date : str
        A string in either ISO, epoch, or predefined format

    Returns
    -------
    datetime.datetime
        Python datetime format.

    Raises
    ------
    ValueError
        Raised when none of the formats are right.

    """
    num_format = re.compile("^[\-]?[1-9][0-9]*\.?[0-9]+$")

    # all the formats the date might be in
    POSSIBLE_DATE_FORMATS = ['%m/%d/%Y', '%Y/%m/%d', "%Y-%m-%dT%H:%M:%S.%fZ"]

    for date_format in POSSIBLE_DATE_FORMATS:
        try:
            return datetime.strptime(date, date_format)  # try to get the date
        except ValueError:
            pass  # if incorrect format, keep trying other formats

    # Check if it's a usable epoch format.
    if num_format.search(date):
        try:
            return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(float(date)))
        except ValueError:
            raise ValueError(f"{date} was not in one of the approved formats.")
    else:
        # Last resort => ISO format
        try:
            return datetime.fromisoformat(date)
        except:
            raise ValueError(f"{date} was not in one of the approved formats.")
    raise ValueError(f"{date} was not in one of the approved formats.")


def filterTimeStamps(relations, filterType: str, **kwargs):
    """Filter dates by their timestamps.

    Parameters
    ----------
    relations : List[Relationship
        List containing the Relationship
        pydantic dataclass. (see dataStructures.py for dataclass definition)
    filterType : str
        Information on how to filter.
        Can be ["betweenWeeks", "inWeek", "betweenDates", "all"]
    kwargs : dict
        Provide {"begin": val, "end": val} in week numbers or
        dates for "betweenWeeks" or "betweenDates" respectively. And provide
        {"week": val} for "inWeek" to filter all in that specific week.
    filterType: str
        How do you want to filter the relations? Possible input is:
        "betweenWeeks" - Provide kwargs["begin], kwargs["end"] in integer weekdays.
        "inWeek" - Provide kwargs["week"] for filtering out a specific week.
        "betweenDates" - Provide kwargs["begin], kwargs["end"] as dates to filter between dates.
        "all" - Get all relations back, no kwargs required.
    **kwargs : str or int
        Input for filter, see filterType for more information.

    Returns
    -------
    List[Relationships, ...]
        List of relationships objects that fall within
        a predefined timeframe.

    """
    if filterType == "betweenWeeks":
        return ([
            i for i in relations
            if kwargs["begin"] < i.week < kwargs["end"]
        ])
    elif filterType == "inWeek":
        return ([
            i for i in relations if kwargs["week"] == i.week
        ])
    elif filterType == "betweenDates":
        return ([
            i for i in relations if extract_date(kwargs["begin"])
            < i.r_timestamp < extract_date(kwargs["end"])
        ])
    elif filterType == "all":
        return relations



"""Visualization helper functions and graphing."""

from matplotlib.ticker import AutoMinorLocator, MultipleLocator


def setStandardFrame(
    ax,
    x_major_td=0,
    x_minor_td=0,
    y_major_td=0,
    y_minor_td=0,
    x_lim=0,
    y_lim=0,
    lw=2,
    major_alpha=0.9,
    minor_alpha=0.7,
):
    """Implement helper function to standardize matplotlib graphs.

    Parameters
    ----------
    ax : matplotlib axis object
        Standard matplotlib axis object.
        E.g., as from output of `fig, ax = plt.subplots()`
    x_minor_td : int
        Set grid minor tick distance.
        This is in relation to the major grid. So a value of 2 result in
        there being 1 extra minor gridline halfway the major lines. 3, means
        at one-third, and two-thirds a minor line, etc. Defaults to 0.
    x_major_td : int
        Set grid major tick distance. Sets major
        gridlines on multiples of this value. Defaults to 0.
    y_major_td : int
        Set locators for grid. See x_minor_td.
        Defaults to 0.
    y_minor_td : int
        Set locators for grid. See x_major_td.
        Defaults to 0.
    x_lim : int
        Viewport limit on x-axis. Defaults to 0.
    y_lim : int
        Viewport limit on y-axis. Defaults to 0.
    lw : int
        Line width of the grid. Defaults to 2.
    major_alpha : float
        Set alpha of major gridlines. Defaults to 0.9.
    minor_alpha : float
        Set alpha of minor gridlines. Defaults to 0.7.

    Returns
    -------
    None

    """
    # Remove ticks
    ax.yaxis.set_tick_params(which="both", length=0)
    ax.xaxis.set_tick_params(which="both", length=0)

    # Customize domain limit if wanted
    if x_lim != 0:
        ax.set_x_lim(x_lim)
    if y_lim != 0:
        ax.set_y_lim(y_lim)

    # Customize minor and major ticks if wanted
    # Minor = majorTickDistance/minorTickDistance
    if x_major_td != 0:
        ax.xaxis.set_major_locator(MultipleLocator(x_major_td))
    if x_minor_td != 0:
        ax.xaxis.set_minor_locator(AutoMinorLocator(x_minor_td))
    if y_major_td != 0:
        ax.yaxis.set_major_locator(MultipleLocator(y_major_td))
    if y_minor_td != 0:
        ax.yaxis.set_minor_locator(AutoMinorLocator(y_minor_td))

    # Turn grid on for both major and minor ticks and style minor slightly
    # differently.
    ax.grid(which="major", color="#CCCCCC", linestyle="--", lw=lw, alpha=major_alpha)
    ax.grid(which="minor", color="#CCCCCC", linestyle=":", lw=lw, alpha=minor_alpha)

    # Remove box around figure
    for spine in ("top", "right", "bottom", "left"):
        ax.spines[spine].set_visible(False)


def setLabelsAndTitles(
    ax, title, xlabel, ylabel, tick_size=20, label_size=20, title_size=22
):
    """Set the labels and titles of the graph including it's (tick)sizes.

    Parameters
    ----------
    ax : matplotlib axis object
        Standard matplotlib axis object.
        E.g., as from output of `fig, ax = plt.subplots()`
    title : str
        Set graph title.
    xlabel : str
        Set label for a-axis.
    ylabel : str
        Set label for y-axis.
    tick_size : int
        Set size for the ticks. Defaults to 20.
    label_size : int
        Set size for the axes labels. Defaults to 20.
    title_size : int
        Set size for the graph title. Defaults to 22.

    Returns
    -------
    None

    """
    for tick in ax.xaxis.get_major_ticks():
        tick.label.set_fontsize(tick_size)

    for tick in ax.yaxis.get_major_ticks():
        tick.label.set_fontsize(tick_size)

    ax.set_xlabel(xlabel, fontsize=label_size)
    ax.set_ylabel(ylabel, fontsize=label_size)
    ax.set_title(title, fontsize=title_size)


def setLegend(ax, loc=0, lr_loc=1.05, ud_loc=1, size=16):
    """Set legend for graphs.

    You can set the legend either at a previously specified location
    by setting the loc, or a tuple made from lr_loc and ud_loc.

    Parameters
    ----------
    ax : matplotlib axis object
        Standard matplotlib axis object.
        E.g., as from output of `fig, ax = plt.subplots()`
    loc : int
        The location of the legend.
        The strings 'upper left', 'upper right', 'lower left',
        'lower right' place the legend at the corresponding
        corner of the axes/figure.
        The strings 'upper center', 'lower center', 'center left',
        'center right' place the legend at the center of the
        corresponding edge of the axes/figure.
        The string 'center' places the legend at the center of the
        axes/figure.
        The string 'best' places the legend at the location, among
        the nine locations defined so far, with the minimum
        overlap with other drawn artists. This option can be
        quite slow for plots with large amounts of data; your
        plotting speed may benefit from providing a specific
        location. Defaults to 0.
    lr_loc : float
        Left-right location. Defaults to 1.05.
    ud_loc : int
        Up-down location. Defaults to 1.
    size : int
        Scaling the legend. Defaults to 16.

    Returns
    -------
    None

    """
    if loc == 0:
        ax.legend(bbox_to_anchor=(lr_loc, ud_loc), prop={"size": size})
    else:
        ax.legend(loc=loc, prop={"size": size})



"""Optimization algorithms."""

import numpy as np
import tqdm


def metro_hast(
        tdata, N, simSetting, perturbSetting, simFunc, perturbFunc, scoreFunc
):
    """Metropolis hastings algorithm that samples parameter space.

    Parameters
    ----------
    tdata : np.array
        True dataset on which the score function calculates its value.
    N : int
        Number of iterations the metropolist hastings algorithm performs.
    simSetting : dict
        Dictionary of settings needed for the simulation.
    perturbSetting : dict
        Dictionary of settings needed for the perturbation function.
    simFunc : function
        Function that performs the actual simulation with np.array output.
    perturbFunc : function
        Function that perturbs variable inputs into simulation function.
    scoreFunc : function
        Function that calculates a score from the fit between real data and
        simulated data.

    Returns
    -------
    simSettings : np.array
        Set of variable settings that were sampled.
    allData : np.array
        Generated data from the previous settings.
    scores : np.array
        Scores according to provided score function regarding data fit.
    """
    # start simulation
    allData = [simFunc(**simSetting)[1]]
    scores = [scoreFunc(tdata, allData[-1])]
    simSettings = [simSetting]

    for i in tqdm(range(N)):
        # Perturb parameters slightly
        newSettings = perturbFunc(simSettings[-1].copy(), perturbSetting)
        # Run simulation
        newData = simFunc(**newSettings)[1]
        # Calculate score
        newScore = scoreFunc(tdata, newData)
        # See if score is better than the last iteration and keep if it is
        if newScore <= scores[-1]:
            simSettings.append(newSettings)
            allData.append(newData)
            scores.append(newScore)
        # Or else give a random chance to still be accepted
        else:
            random = np.random.rand()
            if random < scores[-1] / newScore:
                simSettings.append(newSettings)
                allData.append(newData)
                scores.append(newScore)
    return np.array(simSettings), np.array(allData), np.array(scores)



"""Simulation bases."""

from scipy.integrate import odeint
import numpy as np


def ODESimulation(ODE, timesteps, resolution, y0, a, f):
    """Set up an ODE simulation.

    Parameters
    ----------
    ODE : Function
        Your function describing the set of ordinary differential equations.
    timesteps : int
        The number of steps you want to simulate.
    resolution : int
        How many datapoints do you want per time step?
    y0 : Tuple
        A tuple of your initial conditions.
    a : Tuple
        A tuple of your variables that you'd eventually want to fit.
    f : Tuple
        A tuple containing the functions you'd want to fit.

    Returns
    -------
    time : Array
        x-axis time data
    data : numpy.ndarray
        Numpy array of your output

    """
    # Set up time
    time = np.linspace(1, timesteps, timesteps * resolution)

    # Get data
    data = odeint(ODE, y0, time, args=(a, f))

    return time, data



"""All commonly used pydantic network dataclasses ready for reuse."""

from typing import List, Optional
from datetime import datetime
import time
import re

from pydantic import BaseModel, validator, Extra

import numpy as np


class Relationship(BaseModel):
    """Dataclass for a single relationship."""

    r_from: str
    r_to: str
    r_timestamp: datetime
    r_type: Optional[str]
    r_strength: Optional[int]

    class Config:
        """Configure dataclass."""

        extra = Extra.allow

    @validator("r_timestamp", pre=True)
    def time_validate(cls, v):
        """Make sure that the date type formatting works out.

        Parameters
        ----------
        v :
            

        Returns
        -------

        """
        num_format = re.compile("^[\-]?[1-9][0-9]*\.?[0-9]+$")

        # all the formats the date might be in
        POSSIBLE_DATE_FORMATS = ['%m/%d/%Y', '%Y/%m/%d', "%Y-%m-%dT%H:%M:%S.%fZ"]

        for date_format in POSSIBLE_DATE_FORMATS:
            try:
                return datetime.strptime(v, date_format)  # try to get the date
            except ValueError:
                pass  # if incorrect format, keep trying other formats

        # Check if it's numerical
        if num_format.search(v):
            try:
                return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(float(v)))
            except ValueError:
                raise ValueError(f"{v} was not in one of the approved formats.")
        else:
            # Last resort => ISO format
            try:
                return datetime.fromisoformat(v)
            except ValueError:
                raise ValueError(f"{v} was not in one of the approved formats.")
        raise ValueError(f"{v} was not in one of the approved formats.")

    @property
    def week(self):
        """ """
        return self.r_timestamp.isocalendar().week

    @property
    def year(self):
        """ """
        return self.r_timestamp.isocalendar().year


class RelationshipData(BaseModel):
    """Encoding all relationships in a dataset."""

    relations: List[Relationship]

    @property
    def active_weeks(self):
        """Get distinct weeks of activity."""
        return np.unique([m.week for m in self.relations])

    @property
    def active_years(self):
        """Get distinct weeks of activity."""
        return np.unique([m.year for m in self.relations])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['researchhelper',
 'researchhelper.analyse',
 'researchhelper.datastructures',
 'researchhelper.decorators',
 'researchhelper.modeling',
 'researchhelper.optimize',
 'researchhelper.processdata',
 'researchhelper.visualize']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4.0,<4.0.0',
 'networkx>=2.6.2,<3.0.0',
 'numpy>=1.20.1,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'scipy>=1.6.2,<2.0.0',
 'tqdm>=4.62.1,<5.0.0']

setup_kwargs = {
    'name': 'researchhelper',
    'version': '0.1.7',
    'description': 'A repository of useful functions supporting research.',
    'long_description': '# Table of Contents\n\n1.  [Welcome to ResearchHelper](#orga7ce240)\n2.  [Sphinx, Pypi](#org6b44add)\n\n\n<a id="orga7ce240"></a>\n\n# Welcome to ResearchHelper\n\nThis is a personal repository of functions that I use throughout projects. Besides that, it could obviously also be useful to others! So see if it holds something for you.\n\n\n<a id="org6b44add"></a>\n\n# Sphinx, Pypi\n\nThis package is not only for my functionality, but also to practice in making a package and documenting is correctly. For that I\'m using Sphinx. The online hosted documentation is found on my website <https://www.baschatel.nl/ResearchHelper>.\n',
    'author': 'Bas Chatel',
    'author_email': 'bastiaan.chatel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.baschatel.nl/researchhelper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Possible functional forms that can be used in system dynamics modelling."""

import numpy as np


def sigmoid(x, height=1, base=0, shift=0.5, slope=10):
    """Sigmoidal function that, using default values is bounded between [0,1] for both axes.

    Parameters
    ----------
    x : int
        Input for function.
    height : float
        Controls the height of the sigmoid. (Default value = 1).
    base : float
        Controls the lowest point of the sigmoid. (Default value = 0)
    shift : float
        Controls the shifting of the sigmoid (to left or right), the midpoint
        will end up here. (Default value = 0.5)
    slope : float
        Controls the steepness or slope of the sigmoid. (Default value = 10)

    Returns
    -------
    value : float
        Calculated y-coordinate of sigmoid.

    """
    return base + (height * (1 / (1 + np.exp((-x + shift) * slope))))


def linear(x, a=1, b=0):
    """Linear function.

    Parameters
    ----------
    x : float
        Input for function.
    a : float
        Slope of the line. (Default value = 1)
    b : float
        The y-intercept. (Default value = 0)

    Returns
    -------
    value : float
        Calculated y-coordinate of linear function.
    """
    return a * x + b
